//******************************************************************************
//
// Copyright (c) 2010 Advantech Industrial Automation Group.
//
// advantech watch dog driver for F7511R chip
// 
// This program is free software; you can redistribute it and/or modify it 
// under the terms of the GNU General Public License as published by the Free 
// Software Foundation; either version 2 of the License, or (at your option) 
// any later version.
// 
// This program is distributed in the hope that it will be useful, but WITHOUT 
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
// FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for 
// more details.
// 
// You should have received a copy of the GNU General Public License along with
// this program; if not, write to the Free Software Foundation, Inc., 59 
// Temple Place - Suite 330, Boston, MA  02111-1307, USA.
// 
//
//
//******************************************************************************
/*
 *
 * File:        advwdt7511.c
 * Version:     1.11.1 <07/10/2006>
 * Author:      jianfeng
 *
 * Description:  *  There are two ways of WDT unit: second and minute.The default unit is second.
 *				They are selected by the bit 3 of F75111R_TIMER_CNT.This program use the default 
 *				unit.The watchdog driver is for F75111R chip to operate Advantech UNO-2172/2182 
 *				hardwares.
 *
 * Status:      working
 *
 * Change Log: 
 *            Version 1.00 <03/08/2006> WangMao        
 *                 - Initial version  
 *            Version 1.01 <07/10/2006> WangMao
 *                 - Add better support for kernel parameter on both 2.4 and 2.6 kernel.
 *            Version 1.10 <09/20/2006> WangMao
 *                 - Add IO address detection (0x2E or 0x4E?)
 *            Version 1.11 <04/01/2009> jianfeng
 *                 - to support FC8,FC9,openSuse etc.
 *            Version 1.12 <05/06/2010> jianfeng
 *      	     to support FC10,FC11,FC12,SUSE 11.2 Madriva 2010
 * ----------------------------------------------------------------------------
 */

#include <linux/version.h>
#define advwdt_SRC
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 18)
#include <linux/config.h>
#endif
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/watchdog.h>
#include <linux/fcntl.h>
#include <linux/version.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/i2c.h>
#include <linux/fs.h>
#include <linux/types.h>
#include <linux/delay.h>
#include <linux/module.h>
#include <linux/smp_lock.h>

#include <linux/miscdevice.h>
#include <linux/notifier.h>
#include <linux/reboot.h>

#include <asm/io.h>
#include <asm/uaccess.h>
#include <asm/system.h>


#define DEVICE_NAME "watchdog"

#if defined(CONFIG_SMP)
#define __SMP__
#endif

#define ADVANTECH_WDT_VER        "1.12"
#define ADVANTECH_WDT_DATE       "05/06/2010"

#define WATCHDOG_MINOR	130

//#define ADVANTECH_DEBUG
#ifdef ADVANTECH_DEBUG
#define DEBUGPRINT     printk
#else
#define DEBUGPRINT(a, ...)
#endif

#ifdef CONFIG_WATCHDOG_NOWAYOUT
static int nowayout = 1;
#else
static int nowayout = 0;
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
MODULE_PARM(nowayout,"i");
#else
module_param(nowayout, int, 0);
#endif
MODULE_PARM_DESC(nowayout, "Watchdog cannot be stopped once started (default=CONFIG_WATCHDOG_NOWAYOUT)");

__u8 registe;
struct i2c_client *new_client;

static unsigned long advwdt_is_open;

static char adv_expect_close;

static int timeout_index = 5;

#ifndef I2C_NAME_SIZE
#define I2C_NAME_SIZE 32
#endif

#ifndef MODULE
static int __init adv_setup(char *str)
{
	int ints[4];

	str = get_options(str, ARRAY_SIZE(ints), ints);

	if(ints[0] > 0){
		advwdt_base_addr = ints[1];
		if(ints[0] > 1)
			advwdt_base_addr = ints[2];
	}
	return 1;
}

__setup("advwdt=", adv_setup);
#endif /* !MODULE */


/* Addresses to scan */
static unsigned short normal_i2c[] ={0X4e,I2C_CLIENT_END };/*0x9c/2=0x4e is device address*/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,31)
static unsigned short normal_i2c_range[] = {0X4e,0X4e,I2C_CLIENT_END };
/* Insmod parameters */
I2C_CLIENT_INSMOD;
static int advwdt_attach_adapter(struct i2c_adapter *adapter);
static int advwdt_detect(struct i2c_adapter *adapter, int address, int kind);
static int advwdt_detach_client(struct i2c_client *client);
#else
I2C_CLIENT_INSMOD_1(advwdt);
static int advwdt_remove(struct i2c_client *client);
static int advwdt_probe(struct i2c_client *client,
			 const struct i2c_device_id *id);
static int advwdt_detect(struct i2c_client *client, int kind,
			  struct i2c_board_info *info);
#endif



/* This is the driver that will be inserted */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 31)
static struct i2c_driver advwdt_driver = {
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,9)
	.name = "advwdt",
#else	
	.driver =
   	{
    		.name = "advwdt",
   	},
#endif

	//.flags = I2C_DF_NOTIFY,
	.attach_adapter = advwdt_attach_adapter,
	.detach_client = advwdt_detach_client,
};
#else
static const struct i2c_device_id advwdt_id[] = {
	{ "advwdt", 1 },
	{ }
};

MODULE_DEVICE_TABLE(i2c,advwdt_id);

static struct i2c_driver advwdt_driver= {
	.class =I2C_CLASS_HWMON,
	.driver = {
		.name	= "advwdt",
	},
	.probe		= advwdt_probe,
	.remove		= advwdt_remove,
	.id_table	= advwdt_id,
	.detect		= advwdt_detect,
	.address_data	= &addr_data,
};
#endif

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 31)
static int advwdt_attach_adapter(struct i2c_adapter *adapter)
{
	return i2c_probe(adapter, &addr_data, advwdt_detect);
}


/* This function is called by i2c_probe */
static int advwdt_detect(struct i2c_adapter *adapter, int address, int kind)
{
 	int err = 0;
	const char *client_name = "";
	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE))
		goto exit;

 /* OK. For now, we presume we have a valid client. We now create the
    client structure, even though we cannot fill it completely yet. */
	//if (!(data = kmalloc(sizeof(struct wdt_data), GFP_KERNEL)))
	if (!(new_client = kmalloc(sizeof(struct i2c_client),GFP_KERNEL)))
	{
 		err = -ENOMEM;
		goto exit;
	}
	memset(new_client,0,sizeof(struct i2c_client));
	
	new_client->addr = address;
	//printk("address==%x\n",new_client->addr);
	new_client->adapter = adapter;
	new_client->driver = &advwdt_driver;
	new_client->flags = 0;

	client_name = "advwdt";

	/* Fill in the remaining client fields and put it into the global list */
	//strlcpy(new_client->name,client_name,I2C_NAME_SIZE);
	strcpy(new_client->name,client_name);
	//printk("client's name is %s in tvp5146_detect\n",new_client->name);

	/* Tell the I2C layer a new client has arrived */
	if ((err = i2c_attach_client(new_client)))
	 	goto exit_free;

	//printk("client success \n");
	return 0;

	/* OK, this is not exactly good programming practice, usually. But it is
   	very code-efficient in this case. */

      exit_free:
     		kfree(new_client);
      exit:
		return err;
}

static int advwdt_detach_client(struct i2c_client *client)
{

	int err;

	if ((err = i2c_detach_client(client))){
		printk("i2c_detach_client\n");
		return err;}

	//kfree(i2c_get_clientdata(client));
	kfree(client);
	
	return 0;
}
#else
static int advwdt_remove(struct i2c_client *client)
{
	return 0;
}

static int advwdt_probe(struct i2c_client *client,
			 const struct i2c_device_id *id)
{
	new_client =client;
	return 0;
}

/* Return 0 if detection is successful, -ENODEV otherwise */
static int advwdt_detect(struct i2c_client *client, int kind,
			  struct i2c_board_info *info)
{
	struct i2c_adapter *adapter = client->adapter;
	if (!i2c_check_functionality(adapter, I2C_FUNC_SMBUS_BYTE_DATA))
		return -ENODEV;
	strlcpy(info->type, "advwdt", I2C_NAME_SIZE);

	return 0;
}
#endif

static int advwdt_wdtaccess ( int offset )
{
	struct i2c_client *client = new_client;
	if(0==offset){
		//Disable WDT
		registe = 0x36;
		i2c_smbus_write_byte_data(client,registe,0x00); 
		return -1;
	}else{
	
		//enable WDT
		registe = 0x36;
		i2c_smbus_write_byte_data(client,registe,0x32); 

		//set WDT time
		registe = 0x37;
		i2c_smbus_write_byte_data(client,registe,offset); 

		return 0;
	}
}


static int advwdt_open( struct inode * inode, struct file * filp)
{ 
	struct i2c_client *client;
	
	if (test_and_set_bit(0, &advwdt_is_open))
		return -EBUSY;

	//printk("new :%X\n",new_client); 
 	client=new_client;
 	client->adapter = new_client->adapter;
 	filp->private_data = client;
	
	//config WDT
 	registe = 0x01;
	i2c_smbus_write_byte_data(client,registe,0x20);

 	registe = 0x03;
	//i2c_smbus_write_byte_data(client,registe,0x03);
	i2c_smbus_write_byte_data(client,registe,0x1B);

	//config WDT control
	registe = 0x36;
	i2c_smbus_write_byte_data(client,registe,0x50);
	/*
	 *	Activate
	 */
	advwdt_wdtaccess( timeout_index );
	return 0;
}

static int 
advwdt_release(struct inode * inode, struct file * filp)
{
	if( adv_expect_close == 42 )
	{
		advwdt_wdtaccess( 0 );
	} else {
		printk(KERN_CRIT "Advantech WDT: Unexpected close, not stopping watchdog!\n");
		advwdt_wdtaccess( timeout_index );
	}
	clear_bit( 0,&advwdt_is_open );
	adv_expect_close = 0;
	return 0;
}


static ssize_t
advwdt_write(struct file *file, const char *buf, size_t count, loff_t *ppos)
{
	/*  Can't seek (pwrite) on this device  */
	// Joshua mask to support 2.6.8 kernel @04/19/2005
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,8)
	if (ppos != &file->f_pos)
		return -ESPIPE;
#endif
	if (count) 
	{
		if (!nowayout) 
		{
			size_t i;
			adv_expect_close = 0;
			for (i = 0; i != count; i++) 
			{
				char c;
				if(get_user(c, buf+i))
					return -EFAULT;
				if (c == 'V')
					adv_expect_close = 42;
			}
		}
		advwdt_wdtaccess( timeout_index );
	}
	return count;
}

static int
advwdt_ioctl(struct inode *inode, struct file *file, unsigned int cmd,
	  unsigned long arg)
{
	unsigned long new_timeout;
	static struct watchdog_info ident = {
// Joshua mask to support 2.4.18 kernel @04/19/2005
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,4,20)
		options:		WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT,/* | WDIOF_MAGICCLOSE,*/ 
#else
		options:		WDIOF_KEEPALIVEPING | WDIOF_SETTIMEOUT | WDIOF_MAGICCLOSE, 

#endif
		firmware_version:	0,
		identity:		"Advantech WDT"
	};

	switch ( cmd ) 
   	{
	case WDIOC_GETSUPPORT:
	   if (copy_to_user((struct watchdog_info *)arg, &ident, sizeof(ident)))
	      return -EFAULT;
	   break;

	case WDIOC_GETSTATUS:
	case WDIOC_GETBOOTSTATUS:
	   return put_user(0, (int *)arg);

	case WDIOC_KEEPALIVE:
	   advwdt_wdtaccess( timeout_index );
	   break;

	case WDIOC_SETTIMEOUT:
	   if (get_user( new_timeout, (unsigned long *)arg))
		  return -EFAULT;
          timeout_index =(int)new_timeout;           
           
          if ( timeout_index <= 0 )
           { 
              advwdt_wdtaccess( 0 );
              return -EINVAL;
           }
           else
           {
	      advwdt_wdtaccess( timeout_index );
           }
	   break;
        
	case WDIOC_GETTIMEOUT: 
    	   if ( timeout_index == 0 )
           {
              return -EFAULT;
           }
        return put_user( timeout_index, (unsigned long *)arg);
        
	case WDIOC_SETOPTIONS:
	{
	  int options, retval = -EINVAL;
	  if (get_user(options, (int *)arg))
	    return -EFAULT;

	  if (options & WDIOS_DISABLECARD) 
          {
             if (!nowayout)
	     {
                advwdt_wdtaccess( 0 );
                printk("WDT disabled.\n");
             }
             else
             {
                printk("CONFIG_WATCHDOG_NOWAYOUT is enabled. WDT can't be disabled.\n");
                return -EFAULT;
             }
	     retval = 0;
	  }

	  if (options & WDIOS_ENABLECARD) {            
            if ( timeout_index == 0 )
               timeout_index = 5;  /* set to the 5 default value */
	    advwdt_wdtaccess( timeout_index );
            DEBUGPRINT( "Enable the watchdog!\n The timeout index is: %d\n", timeout );
	    retval = 0;
	  }
	  return retval;      
	}
        break;

	default:
	  return -ENOTTY;
	}
	return 0;
}

/*
 *	Notifier for system down
 */
static int
advwdt_notify_sys(struct notifier_block *this, unsigned long code,
	void *unused)
{
	if (code == SYS_DOWN || code == SYS_HALT) {
		/* Turn the WDT off */
		advwdt_wdtaccess( 0 );
	}
	return NOTIFY_DONE;
}

/*to user*/
static struct file_operations advwdt_fops = {
	.owner = THIS_MODULE,
	.open=advwdt_open,
  	.write = advwdt_write,
	.release = advwdt_release,
	.ioctl = advwdt_ioctl,
};

static struct miscdevice advwdt_miscdev = {
	minor:		WATCHDOG_MINOR,
	name:		"watchdog",
	fops:		&advwdt_fops,
};
/*
 *	The WDT needs to learn about soft shutdowns in order to
 *	turn the timebomb registers off.
 */ 
static struct notifier_block advwdt_notifier = {
	advwdt_notify_sys,
	NULL,
	0
};

static int __init advwdt_init(void)
{
	int ret,result;
 
	result = misc_register(&advwdt_miscdev);
	register_reboot_notifier(&advwdt_notifier);
	if(result)
 	{
  		printk(KERN_ERR DEVICE_NAME ": Unable to get major %d\n", 10);
  		return(result);
	}    
	ret=i2c_add_driver(&advwdt_driver);
	if (ret)
      		goto out_unreg_class;

	printk("Watchdog Timer first disabled\n");

        printk( "========================================================\n");
	printk( "         WDT driver for F75111R Chip \n");
        printk( "                   V%s [%s]\n", 
              ADVANTECH_WDT_VER, ADVANTECH_WDT_DATE);
        printk( "            Advantch eAutomation Devision.\n");
        printk( "========================================================\n");
        printk( "CONFIG_WATCHDOG_NOWAYOUT is currently %s\n", nowayout?"enabled!":"disabled!");

	// printk(KERN_ERR DEVICE_NAME ": init OK\n");
	return 0;
	out_unreg_class:
 	misc_deregister(&advwdt_miscdev);
	return -1;
}

static void __exit advwdt_exit(void)
{
	i2c_del_driver(&advwdt_driver);
	misc_deregister(&advwdt_miscdev);
	unregister_reboot_notifier(&advwdt_notifier);
}


module_init(advwdt_init);
module_exit(advwdt_exit);

MODULE_DESCRIPTION("Advantech WDT driver");
MODULE_LICENSE("GPL");
